﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Windows;
using System.Windows.Controls;
using System.Windows.Input;
using System.Windows.Media;

using AudioSwitcher.AudioApi;
using AudioSwitcher.AudioApi.CoreAudio;
using AudioSwitcher.AudioApi.Observables;

using NAudio.CoreAudioApi;
using NAudio.Wave;

namespace TestProject
{
    /// <summary>
    /// 오디오 컨트롤
    /// </summary>
    public partial class AudioControl : UserControl
    {
        //////////////////////////////////////////////////////////////////////////////////////////////////// Field
        ////////////////////////////////////////////////////////////////////////////////////////// Private

        #region Field

        /// <summary>
        /// 코어 오디오 장치 리스트
        /// </summary>
        private List<CoreAudioDevice> coreAudioDeviceList = null;

        /// <summary>
        /// 디폴트 코어 오디오 장치
        /// </summary>
        private CoreAudioDevice defaultCoreAudioDevice = null;

        /// <summary>
        /// 코어 오디오 컨트롤러
        /// </summary>
        private CoreAudioController coreAudioController = null;

        /// <summary>
        /// Disposable 리스트
        /// </summary>
        private List<IDisposable> disposableList = null;

        /// <summary>
        /// 볼륨 변경 액션
        /// </summary>
        private Action<DeviceVolumeChangedArgs> changeVolumeAction = null;

        /// <summary>
        /// 음소거 변경 액션
        /// </summary>
        private Action<DeviceMuteChangedArgs> changeMuteMicrophoneAction = null;

        /// <summary>
        /// 디폴트 장치 변경 액션
        /// </summary>
        private Action<DefaultDeviceChangedArgs> changeDefaultDeviceAction = null;

        /// <summary>
        /// 장치 상태 변경 액션
        /// </summary>
        private Action<DeviceStateChangedArgs> changeDeviceStateAction = null;

        /// <summary>
        /// 멀티미디어 장치 열거자
        /// </summary>
        private MMDeviceEnumerator mmDeviceEnumerator = null;

        /// <summary>
        /// 웨이브 입력 이벤트
        /// </summary>
        private WaveInEvent waveInEvent = null;

        /// <summary>
        /// 중단 여부
        /// </summary>
        private bool stopped = false;

        /// <summary>
        /// 최대 오디오 값
        /// </summary>
        private double maximumAudioValue = 0;

        /// <summary>
        /// 마지막 오디오 값
        /// </summary>
        private double lastAudioValue = 0;

        /// <summary>
        /// 샘플 비율
        /// </summary>
        private const int SAMPLE_RATE = 44100;

        /// <summary>
        /// 버퍼 샘플 수
        /// </summary>
        private const int BUFFER_SAMPLE_COUNT = 1024;

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Constructor
        ////////////////////////////////////////////////////////////////////////////////////////// Public

        #region 생성자 - AudioControl()

        /// <summary>
        /// 생성자
        /// </summary>
        public AudioControl()
        {
            InitializeComponent();

            Loaded                                   += UserControl_Loaded;
            Unloaded                                 += UserControl_Unloaded;
            this.microphoneComboBox.SelectionChanged += microphoneComboBox_SelectionChanged;
            this.openSoundControlButton.Click        += openSoundControlButton_Click;
            this.muteMicrophoneButton.Click          += muteMicrophoneButton_Click;
            this.volumeSlider.ValueChanged           += volumeSlider_ValueChanged;
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////////////////////////// Method
        ////////////////////////////////////////////////////////////////////////////////////////// Private
        //////////////////////////////////////////////////////////////////////////////// Event

        #region 사용자 컨트롤 로드시 처리하기 - UserControl_Loaded(sender, e)

        /// <summary>
        /// 사용자 컨트롤 로드시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void UserControl_Loaded(object sender, RoutedEventArgs e)
        {
            this.microphoneComboBox.Cursor = Cursors.Hand;

            this.muteMicrophoneButton.Cursor = Cursors.Hand;

            this.volumeSlider.Cursor  = Cursors.Hand;
            this.volumeSlider.Maximum = 100;
            this.volumeSlider.Value   = 0;

            this.volumeProgressBar.Value = 0;

            this.openSoundControlButton.Cursor = Cursors.Hand;

            this.changeVolumeAction         = ChangeVolume;
            this.changeMuteMicrophoneAction = ChangeMuteMicrophone;
            this.changeDefaultDeviceAction  = ChangeDefaultDevice;
            this.changeDeviceStateAction    = ChangeDeviceState;

            this.disposableList = new List<IDisposable>();

            this.mmDeviceEnumerator = new MMDeviceEnumerator();

            this.coreAudioController = new CoreAudioController();

            SetCoreAudioDeviceList();

            SetControl();
        }

        #endregion
        #region 사용자 컨트롤 언로드시 처리하기 - UserControl_Unloaded(sender, e)

        /// <summary>
        /// 사용자 컨트롤 언로드시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void UserControl_Unloaded(object sender, RoutedEventArgs e)
        {
            if(this.defaultCoreAudioDevice != null)
            {
                this.defaultCoreAudioDevice.Dispose();
            }

            if(this.coreAudioController != null)
            {
                this.coreAudioController.Dispose();
            }

            if(this.mmDeviceEnumerator != null)
            {
                this.mmDeviceEnumerator.Dispose();
            }

            if(this.waveInEvent != null)
            {
                this.waveInEvent.Dispose();
            }
        }

        #endregion
        #region 마이크로폰 콤보 박스 선택 변경시 처리하기 - microphoneComboBox_SelectionChanged(sender, e)

        /// <summary>
        /// 마이크로폰 콤보 박스 선택 변경시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void microphoneComboBox_SelectionChanged(object sender, SelectionChangedEventArgs e)
        {
            if(this.microphoneComboBox.SelectedIndex > -1)
            {
                if(!this.defaultCoreAudioDevice.FullName.Equals(this.microphoneComboBox.SelectedItem.ToString()))
                {
                    CoreAudioDevice coreAudioDevice = this.coreAudioDeviceList.Find(x => x.FullName.Equals(this.microphoneComboBox.SelectedItem.ToString()));

                    if(coreAudioDevice != null)
                    {
                        this.coreAudioController.SetDefaultDevice(coreAudioDevice);
                    }
                }
            }
        }

        #endregion
        #region 사운드 컨트롤 열기 버튼 클릭시 처리하기 - openSoundControlButton_Click(sender, e)

        /// <summary>
        /// 사운드 컨트롤 열기 버튼 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void openSoundControlButton_Click(object sender, RoutedEventArgs e)
        {
            // 제어판/소리/녹음 탭을 연다(마지막 0은 재생 탭, 1은 녹음 탭)
            Process.Start("rundll32", "Shell32.dll,Control_RunDLL Mmsys.cpl,,1");
        }

        #endregion
        #region 마이크로폰 음소거 버튼 클릭시 처리하기 - muteMicrophoneButton_Click(sender, e)

        /// <summary>
        /// 마이크로폰 음소거 버튼 클릭시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void muteMicrophoneButton_Click(object sender, RoutedEventArgs e)
        {
            if(this.defaultCoreAudioDevice != null)
            {
                this.defaultCoreAudioDevice.Mute(!this.defaultCoreAudioDevice.IsMuted);
            }
        }

        #endregion
        #region 볼륨 슬라이더 값 변경시 처리하기 - volumeSlider_ValueChanged(sender, e)

        /// <summary>
        /// 볼륨 슬라이더 값 변경시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void volumeSlider_ValueChanged(object sender, RoutedPropertyChangedEventArgs<double> e)
        {
            if(this.defaultCoreAudioDevice != null)
            {
                this.defaultCoreAudioDevice.Volume = (int)e.NewValue;
            }
        }

        #endregion

        #region 웨이브 입력 이벤트 데이터 이용 가능시 처리하기 - waveInEvent_DataAvailable(sender, e)

        /// <summary>
        /// 웨이브 입력 이벤트 데이터 이용 가능시 처리하기
        /// </summary>
        /// <param name="sender">이벤트 발생자</param>
        /// <param name="e">이벤트 인자</param>
        private void waveInEvent_DataAvailable(object sender, WaveInEventArgs e)
        {
            float maximum = 0;

            for(int i = 0; i < e.BytesRecorded; i += 2)
            {
                short sample = (short)((e.Buffer[i + 1] << 8) | e.Buffer[i + 0]);

                float sampleFloat = sample / 32768f;

                if(sampleFloat < 0)
                {
                    sampleFloat = -sampleFloat;
                }

                if(sampleFloat > maximum)
                {
                    maximum = sampleFloat;
                }
            }

            if(maximum > this.maximumAudioValue)
            {
                this.maximumAudioValue = (double)maximum;
            }

            this.lastAudioValue = maximum;

            double fraction = this.lastAudioValue / this.maximumAudioValue;

            this.InvokeOnUIThreadIfRequired(() => this.volumeProgressBar.Value = (int)(fraction * this.volumeProgressBar.Width));
        }

        #endregion

        //////////////////////////////////////////////////////////////////////////////// Function

        #region Disposable 리스트 지우기 - ClearDisposableList()

        /// <summary>
        /// Disposable 리스트 지우기
        /// </summary>
        private void ClearDisposableList()
        {
            if(this.disposableList != null)
            {
                if(this.disposableList.Count > 0)
                {
                    foreach(IDisposable disposable in this.disposableList)
                    {
                        disposable.Dispose();
                    }
                }

                this.disposableList.Clear();
            }
        }

        #endregion
        #region 코어 오디어 장치 리스트 설정하기 - SetCoreAudioDeviceList()

        /// <summary>
        /// 코어 오디어 장치 리스트 설정하기
        /// </summary>
        private void SetCoreAudioDeviceList()
        {
            if(this.coreAudioDeviceList == null)
            {
                this.coreAudioDeviceList = new List<CoreAudioDevice>();
            }
            else
            {
                this.coreAudioDeviceList.Clear();
            }

            IEnumerable<CoreAudioDevice> deviceEnumerable = this.coreAudioController.GetCaptureDevices().Where
            (
                device => !device.FullName.Contains("Unknown") && device.State.Equals(AudioSwitcher.AudioApi.DeviceState.Active)
            );

            foreach(CoreAudioDevice coreAudioDevice in deviceEnumerable)
            {
                this.coreAudioDeviceList.Add(coreAudioDevice);
            }
                
            if(this.coreAudioDeviceList.Count > 0)
            {
                ClearDisposableList();

                this.defaultCoreAudioDevice = this.coreAudioController.DefaultCaptureDevice;

                for(int i = 0; i < this.coreAudioDeviceList.Count; i++)
                {
                    CoreAudioDevice device = this.coreAudioDeviceList[i];

                    this.disposableList.Add(ObservableExtensions.Subscribe(device.VolumeChanged , changeVolumeAction        ));
                    this.disposableList.Add(ObservableExtensions.Subscribe(device.MuteChanged   , changeMuteMicrophoneAction));
                    this.disposableList.Add(ObservableExtensions.Subscribe(device.DefaultChanged, changeDefaultDeviceAction ));
                    this.disposableList.Add(ObservableExtensions.Subscribe(device.StateChanged  , changeDeviceStateAction   ));
                }
            }
        }

        #endregion
        #region 장치 인덱스 구하기 - GetDeviceIndex()

        /// <summary>
        /// 장치 인덱스 구하기
        /// </summary>
        /// <returns>장치 인덱스</returns>
        private int? GetDeviceIndex()
        {
            MMDevice device = this.mmDeviceEnumerator.GetDefaultAudioEndpoint
            (
                DataFlow.Capture,
                NAudio.CoreAudioApi.Role.Multimedia
            );

            int? deviceIndex = 0;

            if(device != null)
            {
                for(int i = 0; i < WaveIn.DeviceCount; i++)
                {
                    if(device.FriendlyName.Equals(WaveIn.GetCapabilities(i).ProductName))
                    {
                        deviceIndex = i;

                        break;
                    }
                }
            }
            else
            {
                return null;
            }

            return deviceIndex;
        }

        #endregion
        #region 웨이브 입력 이벤트 설정하기 - SetWaveInEvent(waveInEvent)

        /// <summary>
        /// 웨이브 입력 이벤트 설정하기
        /// </summary>
        /// <param name="waveInEvent">웨이브 입력 이벤트</param>
        private void SetWaveInEvent(ref WaveInEvent waveInEvent)
        {
            int? deviceIndex = GetDeviceIndex();

            if(deviceIndex != null)
            {
                waveInEvent = new WaveInEvent();

                waveInEvent.DeviceNumber       = (int)deviceIndex;
                waveInEvent.WaveFormat         = new WaveFormat(44100, 16, 1);
                waveInEvent.BufferMilliseconds = (int)((double)BUFFER_SAMPLE_COUNT / (double)SAMPLE_RATE * 1000.0);

                waveInEvent.DataAvailable    += waveInEvent_DataAvailable;
                waveInEvent.RecordingStopped += (sender, e) => this.stopped = false;
                
                this.InvokeOnUIThreadIfRequired(() => this.volumeProgressBar.Value = 0);
            }
        }

        #endregion
        #region 컨트롤 설정하기 - SetControl()

        /// <summary>
        /// 컨트롤 설정하기
        /// </summary>
        private void SetControl()
        {
            try
            {
                lock(this.microphoneComboBox)
                {
                    this.InvokeOnUIThreadIfRequired(() => this.microphoneComboBox.Items.Clear());

                    if(this.coreAudioDeviceList != null && this.coreAudioDeviceList.Count > 0 && this.defaultCoreAudioDevice != null)
                    {
                        Action<int> action = (count) =>
                        {
                            for(int i = 0; i < count; i++)
                            {
                                this.microphoneComboBox.Items.Add(this.coreAudioDeviceList[i].FullName);
                            }

                            this.microphoneComboBox.SelectedIndex = this.coreAudioDeviceList.IndexOf(this.defaultCoreAudioDevice);
                        };

                        this.InvokeOnUIThreadIfRequired(() => action(this.coreAudioDeviceList.Count));

                        if(this.waveInEvent == null)
                        {
                            SetWaveInEvent(ref this.waveInEvent);
                        }

                        if((int)this.defaultCoreAudioDevice.Volume > -1)
                        {
                            this.InvokeOnUIThreadIfRequired(() => this.volumeSlider.Value = (int)this.defaultCoreAudioDevice.Volume);
                        }

                        if(this.defaultCoreAudioDevice.IsMuted)
                        {
                            this.InvokeOnUIThreadIfRequired(() => this.muteMicrophoneButton.Background = (SolidColorBrush)new BrushConverter().ConvertFromString("red"));

                            if(this.waveInEvent != null)
                            {
                                this.waveInEvent.StopRecording();

                                this.stopped = false;

                                this.InvokeOnUIThreadIfRequired(() => this.volumeProgressBar.Value = 0);
                            }
                        }
                        else
                        {
                            this.InvokeOnUIThreadIfRequired(() => this.muteMicrophoneButton.Background = (SolidColorBrush)new BrushConverter().ConvertFromString("green"));

                            if(this.waveInEvent != null && !this.stopped)
                            {
                                this.waveInEvent.StartRecording();

                                this.stopped = true;
                            }
                        }
                    }
                    else
                    {
                        this.InvokeOnUIThreadIfRequired(() => this.microphoneComboBox.SelectedIndex = -1);
                    }
                }
            }
            catch(Exception exception)
            {
                MessageBox.Show(exception.ToString());
            }
        }

        #endregion

        #region 볼륨 변경하기 - ChangeVolume(e)

        /// <summary>
        /// 볼륨 변경하기
        /// </summary>
        /// <param name="e">이벤트 인자</param>
        private void ChangeVolume(DeviceVolumeChangedArgs e)
        {
            if(e.Device.FullName.Equals(this.defaultCoreAudioDevice.FullName))
            {
                this.InvokeOnUIThreadIfRequired(() => this.volumeSlider.Value = (int)e.Volume);
            }
        }

        #endregion
        #region 마이크로폼 음소거 변경하기 - ChangeMuteMicrophone(e)

        /// <summary>
        /// 마이크로폼 음소거 변경하기
        /// </summary>
        /// <param name="e">이벤트 인자</param>
        private void ChangeMuteMicrophone(DeviceMuteChangedArgs e)
        {
            if(e.Device.FullName.Equals(this.defaultCoreAudioDevice.FullName))
            {
                SetCoreAudioDeviceList();

                SetControl();
            }
        }

        #endregion
        #region 디폴트 장치 변경하기 - ChangeDefaultDevice(e)

        /// <summary>
        /// 디폴트 장치 변경하기
        /// </summary>
        /// <param name="e">이벤트 인자</param>
        private void ChangeDefaultDevice(DefaultDeviceChangedArgs e)
        {
            if(((CoreAudioDevice)e.Device).IsDefaultDevice)
            {
                if(this.waveInEvent != null)
                {
                    this.waveInEvent.StopRecording();

                    this.waveInEvent.DataAvailable -= waveInEvent_DataAvailable;

                    this.waveInEvent = null;
                }

                SetCoreAudioDeviceList();

                SetControl();
            }
        }

        #endregion
        #region 장치 상태 변경하기 - ChangeDeviceState(e)

        /// <summary>
        /// 장치 상태 변경하기
        /// </summary>
        /// <param name="e">이벤트 인자</param>
        private void ChangeDeviceState(DeviceStateChangedArgs e)
        {
            this.InvokeOnUIThreadIfRequired(() => this.volumeProgressBar.Value = 0);

            if(((CoreAudioDevice)e.Device).IsDefaultDevice)
            {
                if(this.waveInEvent != null)
                {
                    this.waveInEvent.StopRecording();

                    this.waveInEvent.DataAvailable -= waveInEvent_DataAvailable;

                    this.waveInEvent = null;
                }
            }

            SetCoreAudioDeviceList();

            SetControl();
        }

        #endregion
    }
}